module Labyrinth where

import VTree

-- Modelling labyrinths.

-- Our labyrinth is a Variadic trees (or a graph):
type Lab alpha = VTree alpha

-- A path in the tree is a list of node labels
type Path alpha = [alpha]

-- A leaf has no children
isLeaf :: Lab alpha-> Bool
isLeaf (Node a ns) = null ns

-- All possible traversals, depth first via fold.
-- Does not work for graphs (infinite trees), as it will diverge!
dfts' :: Lab alpha-> [Path alpha]
dfts' = foldT add where
  add a [] = [[a]]
  add a ps = concatMap (map (a :)) ps

-- All possible traversals, depth first, with length limit
dfts :: Eq alpha=> (alpha-> Bool)-> Lab alpha-> [Path alpha]
dfts trg = dfts0 [] where
  dfts0 p n@(Node a ns)
    | trg a     = [reverse (a:p)]
    | elem a p  = [] 
    | otherwise = concatMap (dfts0 (a:p)) ns 

-- All possible traversals, breadth first, ordered by length
bfts :: Lab alpha-> [Path alpha]
bfts l = bfts0 [] [l] where
  bfts0 p [] = []
  bfts0 p (Node a cs:ns) =
    reverse (a:p) : (bfts0 p ns ++ bfts0 (a:p) cs)


-- An identifier for the nodes:
type Id  = Int

-- An example labyrinth without cycles:
l00 = Node 0 [l01]
l10 = Node 1 [l00]
l20 = Node 2 [l21]
l30 = Node 3 [l20]
l40 = Node 4 []
l01 = Node 5 [l02 ]
l11 = Node 6 []
l21 = Node 7 [l11]
l31 = Node 8 [l30, l41]
l41 = Node 9 [l40, l42]
l02 = Node 10 [l12, l03]
l12 = Node 11 [l13]
l22 = Node 12 [l32]
l32 = Node 13 [l31]
l42 = Node 14 []
l03 = Node 15 [l04]
l13 = Node 16 [l23]
l23 = Node 17 [l22, l24 ]
-- l23 = Node 17 [l24 -- this change bisects the labyrinth]
l33 = Node 18 [l43]
l43 = Node 19 [l44]
l04 = Node 20 [l14]
l14 = Node 21 []
l24 = Node 22 [l34]
l34 = Node 23 [l33]
l44 = Node 24 []


-- An example labyrinth with cycles:
m00 = Node 0 [m01]
m10 = Node 1 [m00]
m20 = Node 2 [m21]
m30 = Node 3 [m20]
m40 = Node 4 []
m01 = Node 5 [m02] -- Alternatively, Node 5 [m11, m02]
m11 = Node 6 [m01] -- This makes one cycle
m21 = Node 7 [m11]
m31 = Node 8 [m30, m41]
m41 = Node 9 [m40, m42]
m02 = Node 10 [m12, m03]
m12 = Node 11 [m13]
m22 = Node 12 [m32]
m32 = Node 13 [m31]
m42 = Node 14 []
m03 = Node 15 [m04]
m13 = Node 16 [m23]
m23 = Node 17 [m22, m24]
m33 = Node 18 [m43]
m43 = Node 19 [m44]
m04 = Node 20 [m14]
m14 = Node 21 [m24] -- This makes the other cycle
m24 = Node 22 [m34] -- Alternatively, Node 22 [m14, m34]
m34 = Node 23 [m33]
m44 = Node 24 []
